package com.riclage.boardview;

import android.content.Context;
import android.os.Parcel;
import android.os.Parcelable;
import android.text.TextUtils;
import android.util.AttributeSet;
import android.util.Log;
import android.view.MotionEvent;
import android.view.View;

import androidx.annotation.IntDef;

import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;

/**
 * Used by Julio on 30/3/2021
 * Created by Ricardo on 27/01/2016.
 */

//Where the 2-D Array for the Board is handled
public class WordSearchBoardView extends TiledBoardView {

    public static final int DIRECTION_UNKNOWN = -1;
    public static final int DIRECTION_RIGHT_HORIZONTAL = 0; //Right Horizontal -> (x, y+1)
    public static final int DIRECTION_SOUTH_EAST_DIAGONAL = 1; //SouthEast Diagonal \> (x+1, y+1)
    public static final int DIRECTION_DOWN_VERTICAL = 2; ////Down Vertical |> (x+1, y)
    public static final int DIRECTION_SOUTH_WEST_DIAGONAL = 3; /* SouthWest Diagonal </ (x+1,y-1) */
    public static final int DIRECTION_LEFT_HORIZONTAL = 4; //Left Horizontal <- (x,y-1)
    public static final int DIRECTION_NORTH_WEST_DIAGONAL = 5; /*NorthWest Diagonal <\ (x-1,y-1)*/
    public static final int DIRECTION_UP_VERTICAL = 6; //Up Vertical <| (x-1,y)
    public static final int DIRECTION_NORTH_EAST_DIAGONAL = 7; //NorthEast Diagonal /> (x-1,y+1)

    public String[][] letterBoard;

    @Retention(RetentionPolicy.SOURCE)
    @IntDef({DIRECTION_UNKNOWN, DIRECTION_RIGHT_HORIZONTAL, DIRECTION_DOWN_VERTICAL, DIRECTION_SOUTH_EAST_DIAGONAL,DIRECTION_SOUTH_WEST_DIAGONAL,DIRECTION_LEFT_HORIZONTAL,DIRECTION_NORTH_WEST_DIAGONAL,DIRECTION_UP_VERTICAL,DIRECTION_NORTH_EAST_DIAGONAL})
    public @interface Direction {}


    public interface OnWordSelectedListener {
        /**
         * Listener for the clients of this board to tell it whether a selected word is valid or not.
         * If a selected word is valid, the board will highlight it, otherwise it will be discarded.
         * @param selectedWord the selected word
         * @return True if the word is valid and should be kept selected
         */
        boolean onWordSelected(BoardWord selectedWord);
    }

    public interface OnWordHighlightedListener {
      /**
       * Listener for the clients of this board to take action after a word has been successfully highlighted.
       * @param highlightedWord the highlighted word
       */
      void onWordHighlighted(BoardWord highlightedWord);
    }

    private SelectedWord currentSelectedWord;
    private List<SelectedWord> selectedWords;

    private OnWordSelectedListener listener;
    private OnWordHighlightedListener highlightedListener;

    public WordSearchBoardView(Context context) {
        super(context);
        init();
    }

    public WordSearchBoardView(Context context, AttributeSet attrs) {
        super(context, attrs);
        init();
    }

    public WordSearchBoardView(Context context, AttributeSet attrs, int defStyleAttr) {
        super(context, attrs, defStyleAttr);
        init();
    }

    @SuppressWarnings("unused")
    public WordSearchBoardView(Context context, AttributeSet attrs, int defStyleAttr, int defStyleRes) {
        super(context, attrs, defStyleAttr, defStyleRes);
        init();
    }

    private void init() {
        selectedWords = new ArrayList<>();
    }

    @Override
    protected void initBoardTileViews(Context context) {
        for (int i = 0; i < getTileCount(); i++) {
            LetterTileView tileView = new LetterTileView(context);
            tileView.setBackgroundResource(getTileBackgroundDrawableResId());
            addView(tileView);
        }
    }

    //Used for scrambling
    @Override
    public void clearBoard() {
        List<SelectedWord> selectedWordsToClear = new ArrayList<>(this.selectedWords);
        this.selectedWords.clear();

        for (SelectedWord word : selectedWordsToClear) { //sets all selected to false
            updateTiles(word.selectedTiles, false, false);
        }
    }

    /**
     * Sets the given letters on the grid
     * @param letterBoard A 2-d (row-by-col) array containing the letters to set
     */
    public void setLetterBoard(String[][] letterBoard) { //Sets the 2-D Array with tiles that have been already given letters
        if (letterBoard.length != getNumRows()
                || letterBoard[0].length != getNumCols()) {
            setBoardSize(letterBoard.length, letterBoard[0].length);
        }

        int row, col;
        for (int i=0; i < getChildCount(); i++) {
            row = i / getNumCols();
            col = i % getNumCols();

            LetterTileView child = (LetterTileView) getChildAt(i);
            child.setLetter(letterBoard[row][col]); //dimensions
        }
    }

    //Checks to see how long the board should be depending on word length
    public List<BoardWord> generateRandomLetterBoard(List<String> validWords) {
        int boardSize = 4; //BoardSize is how long the word can be on said board and does not have to do with the actual board size, depends on longest word size
        for (String word : validWords) {
            if (word.length() > boardSize) {
                boardSize = word.length();
            }
        }
        //return generateRandomLetterBoard(validWords, Math.min(10, boardSize + 1));//Math makes up the board size?
        return generateRandomLetterBoard(validWords, 10);
    }

    /**IMPORTANT-FillIt IMPLEMENTED HERE******************************************************************************************************/
    //Generates the LetterBoard Array
/* direction:

        0
    7	|	1
6	-	a   -	2
    5	|	3
        4
*/

    public List<BoardWord> generateRandomLetterBoard(List<String> validWords, int boardSize) {
        letterBoard = new String[boardSize][boardSize];
        List<BoardWord> wordLocations = new ArrayList<>(validWords.size());

        int i, x, y, tries; //i, x, y are for the loops, tries counts how many tries it takes to randomly place on the board
        boolean done, in = false, all = false; //done and in are used to signify if the word has successfully been placed so get out of loop, all is used for trying all directions
        //if all is true, that means to randomly place word on board since we tried all other directions on trying to overlap
        @Direction int dir, startdir; //dir is current direction and startdir is the direction that was first used to TRY to place the word
        int MAXTRIES = 100; //has a maximum of 100 tries to prevent from trying to randomly place above 100 times to avoid infinite loops
        int DIRECTIONS = 8; //8 directions
        BoardPoint neoPoint; //to compare characters in the board

        x=0;y=0;
        dir=DIRECTION_RIGHT_HORIZONTAL;
        //Inserts all words first
        for (String word : validWords) {
            if (word.length() > boardSize) { //previous function avoids this by calculating board size
                throw new IllegalArgumentException("Word '" + word + "' is longer than the specified board size");
            }

            Random r = new Random();
            done = false; //personal use
            tries = 0;
            startdir = dir; //starts at the direction that the previous direction calculated randomly
            in = false; all = false; //neither placed on the board nor tried all directions already
            for (x = 0; x < boardSize; x++) //row
            {
                for (y = 0; y < boardSize; y++) //column
                {
                    neoPoint = new BoardPoint(x, y);
                    Log.e("neoPoint", neoPoint.toString());
                    if (letterBoard[neoPoint.row][neoPoint.col]!=null && letterBoard[neoPoint.row][neoPoint.col].equals(String.valueOf(word.charAt(0)))) //Step 1: Check throughout whole board to see if a letter from a word already on the board matches the first letter of the word trying to be placed
                    {
                        Log.e("letterboard", "About to try Directions");
                        in = false;
                        all = false;
                        while (!in && !all) //Until either placed on the board or tried all directions possible
                        {
                            Log.e("letterboard", "In While Loop");
                            if (canInsertWordOnBoard(word, new BoardPoint(x, y), dir, letterBoard))		{ //Step 2: Finds a good fit for the word trying all 8 directions
                                in = true; //signifies that the word has been successfully placed on the board
                                done = true; //personal use
                                //dir = (dir + 1) % 8; //next direction for next word
                            }
                            else	{ //Didn't fit using specified direction, so change direction
                                dir = (dir + 1) % 8; //next direction is remainder of mod
                                if (dir == startdir)
                                    all = true; //have we tried all directions, yes if dir=startdir
                            }
                        } // while
                    }// if matched first char
                    if (in) break;
                } // for y
                if (in) break;
            } // for x
            // here if unable to attach to first char of string found in grid
            // so let's try random placement
            if (!in) //Optional Step 3: Randomly placed word on board if unable to overlap with other words
            {
                Log.e("Random", "About to try random Directions");
                // random try
                tries = 0;
                done = false;
                while (!done && tries < MAXTRIES) //Until word placed or tried 100 times
                {
                    tries++;
                    if (!in)
                    {
                        x = r.nextInt(10) % boardSize; //random number % 10
                        y = r.nextInt(10) % boardSize; //random number % 10
                        in = false;
                        all = false;
                        while (!in && !all)
                        {
                            Log.e("Random", "In While Loop"); //ERROR
                            if (canInsertWordOnBoard(word, new BoardPoint(x, y),dir,letterBoard)) //Tries to place word starting with x and y in direction specified
                            {
                                in = true;
                                done = true;
                                //dir = (dir + 1) % 8; //next direction for next word
                            }
                            else
                            {
                                dir = (dir + 1) % 8;
                                if (dir == startdir)
                                    all = true;
                            }
                        }
                    }
                } // while
                // for
            }

            /**IMPORTANT-Fits Part 3 IMPLEMENTED HERE******************************************************************************************************/
            BoardPoint startPoint = new BoardPoint(x, y); //using row and col the word started at
            BoardPoint currPoint = startPoint;
            for (char c : word.toCharArray()) { //Puts word in specified direction and location
                Log.e("Characters", "About to actually put the word on the board");
                //if statement to avoid error
                letterBoard[currPoint.row][currPoint.col] = String.valueOf(c); //puts character /**BREAKS HERE FOR LONG WORDS********************************************************************************************************/
                Log.e("Characters", "Before Shift");
                currPoint = shift(currPoint, dir, 1); //move in the direction indicated
            }
            Log.e("Successful", "Added word to List");
            wordLocations.add(new BoardWord(word, dir, startPoint)); //Adds Coordinates and Direction to array of words
            //Starts all over again
            dir = (dir + 1) % 8; //next direction for next word
        } //end For Loop for all Words

        /**IMPORTANT-ChangeStars IMPLEMENTED HERE******************************************************************************************************/
        //Fills the rest of the board with lowercase letters
        for (int row = 0; row < letterBoard.length; row++) {
            for (int col = 0; col < letterBoard[row].length; col++) {
                if (TextUtils.isEmpty(letterBoard[row][col])) {
                    Random r = new Random();
                    Log.e("ChangeStars", "About to replace stars with letters");
                    //TODO: This only considers the English alphabet for now
                    //Converted to Uppercase(2021)
                    letterBoard[row][col] = String.valueOf((char) (r.nextInt(26) + 'a')).toUpperCase();
                }
            }
        }
        setLetterBoard(letterBoard); //Sets the view to the letterboard view
        return wordLocations; //Returns words with locations to GamePlay.java
    }

    /**IMPORTANT-Fits IMPLEMENTED HERE******************************************************************************************************/
    //Using the startpoint, hypothetically places the word on the board
    private boolean canInsertWordOnBoard(String word, BoardPoint startPoint, @Direction int selectionType, String[][] letterBoard) {
        BoardPoint currPoint = startPoint;
        int x;
        int y;
        Log.e("InsertWordOnBoard", "About to hypothetically place characters");
        for (char wordLetter : word.toCharArray()) { //for every character in the word, check to see if it is empty and the same letter is already there(Will cause errors if same overlapping letters)
            String boardLetter = letterBoard[currPoint.row][currPoint.col];
            if (!TextUtils.isEmpty(boardLetter) && !boardLetter.equals(String.valueOf(wordLetter))) {
                return false;
            }
            Log.e("InsertWordOnBoardLoop", word);
            Log.e("InsertWordOnBoardLoop", String.valueOf(wordLetter));
            x = currPoint.row;
            y = currPoint.col;
            Log.e("InsertWordOnBoardLoop", String.valueOf(x)+String.valueOf(y));
            if(x+1 < 10 && y+1 < 10 && x-1 > 0 && y-1 > 0) { //This is needed to avoid going out of bounds
                currPoint = shift(currPoint, selectionType, 1); //tries a different direction
            }
            else
            {
                return false;
            }
        }
        return true;
    }

    /**IMPORTANT-Fits Part 1 IMPLEMENTED HERE******************************************************************************************************/
    //Shift is where all different directions are defined
    private BoardPoint shift(BoardPoint point, @Direction int direction, int n) {
        int dx, dy;
        Log.e("Shift Direction", String.valueOf(direction));
        switch (direction) //Depending on the direction that was passed as a parameter, we will either add/subtract 0 or 1 from the coordinates/row,col to place the word
        {
            case DIRECTION_RIGHT_HORIZONTAL: dx = 0; dy = 1; //Right Horizontal -> (Forwards)
                return new BoardPoint(point.row+dx,point.col+dy);
            case DIRECTION_SOUTH_EAST_DIAGONAL: dx = 1; dy = 1; //SouthEast Diagonal \>
                return new BoardPoint(point.row+dx,point.col+dy);
            case DIRECTION_DOWN_VERTICAL: dx = 1; dy = 0; //Down Vertical |> (Up Start to Down End)
                return new BoardPoint(point.row+dx,point.col+dy);
            case DIRECTION_SOUTH_WEST_DIAGONAL: dx = 1; dy = -1; /* SouthWest Diagonal </ */
                return new BoardPoint(point.row+dx,point.col+dy);
            case DIRECTION_LEFT_HORIZONTAL: dx = 0; dy = -1; //Left Horizontal <- (Backwards)
                return new BoardPoint(point.row+dx,point.col+dy);
            case DIRECTION_NORTH_WEST_DIAGONAL: dx = -1; dy = -1; /*NorthWest Diagonal <\ */
                return new BoardPoint(point.row+dx,point.col+dy);
            case DIRECTION_UP_VERTICAL: dx = -1; dy = 0; //Up Vertical <| (Down Start to Up End)
                return new BoardPoint(point.row+dx,point.col+dy);
            case DIRECTION_NORTH_EAST_DIAGONAL: dx = -1; dy = 1; //NorthEast Diagonal />
                return new BoardPoint(point.row+dx,point.col+dy);
        }
        return point;
    }

    public void setOnWordSelectedListener(OnWordSelectedListener listener) {
        this.listener = listener;
    }

    public void setOnWordHighlightedListener(OnWordHighlightedListener listener) {
      this.highlightedListener = listener;
    }

    /**IMPORTANT*******************************************************************************************************************************************************************/
    //Used for Save Game Button
    @Override
    public Parcelable onSaveInstanceState() {
        List<LetterTile> tiles = new ArrayList<>(getChildCount());
        int row, col;
        for (int i = 0; i < getChildCount(); i++) {
            row = i / getNumCols();
            col = i % getNumCols();
            View child = getChildAt(i);
            tiles.add(new LetterTile(row, col, child));
        }

        Parcelable p = super.onSaveInstanceState();
        SavedState savedState = new SavedState(p);
        savedState.boardRows = getNumRows();
        savedState.boardCols = getNumCols();
        savedState.boardTiles = tiles;
        savedState.selectedWords = this.selectedWords;
        return savedState;
    }

    /**IMPORTANT*******************************************************************************************************************************************************************/
    //Restores the SavedGame for Resume Game Button
    @Override
    public void onRestoreInstanceState(Parcelable state) {
        SavedState savedState = (SavedState) state;
        super.onRestoreInstanceState(savedState.getSuperState());
        this.selectedWords = savedState.selectedWords;

        String[][] letterBoard = new String[savedState.boardRows][savedState.boardCols];
        for (LetterTile tile : savedState.boardTiles) {
            letterBoard[tile.row][tile.col] = tile.letter;
        }
        setLetterBoard(letterBoard);

        for (SelectedWord word : selectedWords) {
            for (Tile tile : word.selectedTiles) {
                tile.view = getChildAt(tile.row, tile.col);
                tile.view.setPressed(false);
                tile.view.setSelected(true);
            }
        }
    }

    //Activated when user touches a Tile
    @Override
    public boolean onTouchEvent(MotionEvent event) {
        float X = event.getX();
        float Y = event.getY();
        int row = (int) (Y / getTileSize());
        int col = (int) (X / getTileSize());

        View child = getChildAt(row, col);

        //Exit on invalid touches
        if (event.getActionMasked() != MotionEvent.ACTION_UP
                && (row >= getNumRows()
                || col >= getNumCols()
                || child == null)) {
            return true;
        }

        super.onTouchEvent(event);
        /**IMPORTANT************************************************************************************************************************************************************/
        //Switch Case to indicate what to do when User Moves
        switch (event.getActionMasked()) {
            case MotionEvent.ACTION_DOWN: //Possibly Add to this to Reduce Limitations
            case MotionEvent.ACTION_MOVE:
                Tile currentTile = new Tile(row, col, child);
                if (currentSelectedWord == null) {
                    currentSelectedWord = new SelectedWord(currentTile);
                } else if (!currentTile.equals(currentSelectedWord.lastTile)
                        && currentSelectedWord.isTileValid(currentTile)) {
                    if (!currentSelectedWord.isTileAllowed(currentTile)) {
                        //Clear the status of the old selection
                        updateTiles(currentSelectedWord.selectedTiles, false, false);
                        //If the current tile is valid but not allowed for the current word selection,
                        //start a new selection that matches the first tile
                        currentSelectedWord = new SelectedWord(currentSelectedWord.getInitialTile());
                    }
                    List<Tile> tiles = getTilesBetween(currentSelectedWord.lastTile, currentTile); //Adds tiles between first tile and current tile
                    if (tiles.size() > 0) {
                        currentSelectedWord.addTiles(tiles);
                    }
                }
                updateTiles(currentSelectedWord.selectedTiles, true, false);
                break;
            case MotionEvent.ACTION_UP:
                if (currentSelectedWord != null) {
                    boolean isValidSelection = (listener != null && listener.onWordSelected(currentSelectedWord.toBoardWord()));
                    updateTiles(currentSelectedWord.selectedTiles, false, isValidSelection);
                    if (isValidSelection) {
                        selectedWords.add(currentSelectedWord);
                      if(highlightedListener != null) {
                        highlightedListener.onWordHighlighted(currentSelectedWord.toBoardWord());
                      }
                    }
                    currentSelectedWord = null;
                }
                break;
            default:
                return false;
        }
        return true;
    }

    private boolean isTileSelected(Tile tile, @Direction int direction) {
        for (SelectedWord word : selectedWords) {
            //A selected tile cannot be selected again for the same selection type
            if (direction == DIRECTION_UNKNOWN || word.direction == direction) {
                for (Tile wordTile : word.selectedTiles) {
                    //Check also the previous tile in the same direction to prevent connected
                    //selected tiles for different words from happening
                    if (wordTile.equals(tile) || wordTile.equals(shift(tile, direction, -1))) {
                        return true;
                    }
                }
            }
        }
        return false;
    }

    public View getChildAt(int row, int col) {
        return getChildAt(col + row * getNumCols());
    }

    private void updateTiles(List<Tile> tiles, boolean pressed, boolean selected) {
        for (Tile tile : tiles) {
            tile.view.setPressed(pressed);
            //Keep the tile selected if it belongs to a previously selected word
            tile.view.setSelected(selected || isTileSelected(tile, DIRECTION_UNKNOWN));
        }
    }

    /**
     * Get all the tiles between the start and end, excluding the
     * start tile but including the end one
     */
    private List<Tile> getTilesBetween(Tile startTile, Tile endTile) {
        List<Tile> tiles = new ArrayList<>();
        @Direction int direction = startTile.getDirectionFrom(endTile);
        BoardPoint currPoint = startTile;
        if (direction != DIRECTION_UNKNOWN) {
            while (!currPoint.equals(endTile)) {
                currPoint = shift(currPoint, direction, 1);

                View child = getChildAt(currPoint.row, currPoint.col);
                Tile t = new Tile(currPoint.row, currPoint.col, child);
                if (isTileSelected(t, direction)) {
                    break;
                } else {
                    tiles.add(t);
                }
            }
        }
        return tiles;
    }

    private static class SelectedWord implements Parcelable {
        private @Direction
        int direction = DIRECTION_UNKNOWN;

        private Tile lastTile;
        private List<Tile> selectedTiles;

        public SelectedWord(Tile initialTile) {
            lastTile = initialTile;
            selectedTiles = new ArrayList<>();
            selectedTiles.add(initialTile);
        }

        protected SelectedWord(Parcel in) {
            //noinspection ResourceType
            direction = in.readInt();
            lastTile = in.readParcelable(Tile.class.getClassLoader());
            selectedTiles = in.createTypedArrayList(Tile.CREATOR);
        }

        public static final Creator<SelectedWord> CREATOR = new Creator<SelectedWord>() {
            @Override
            public SelectedWord createFromParcel(Parcel in) {
                return new SelectedWord(in);
            }

            @Override
            public SelectedWord[] newArray(int size) {
                return new SelectedWord[size];
            }
        };

        public boolean isTileValid(Tile tile) {
            return getInitialTile().getDirectionFrom(tile) != DIRECTION_UNKNOWN;
        }

        public boolean isTileAllowed(Tile tile) {
            @Direction int currType = lastTile.getDirectionFrom(tile);
            return currType != DIRECTION_UNKNOWN
                    && (direction == DIRECTION_UNKNOWN || direction == currType);
        }

        public Tile getInitialTile() {
            return selectedTiles.get(0);
        }

        public void addTiles(List<Tile> tiles) {
            if (direction == DIRECTION_UNKNOWN) {
                direction = lastTile.getDirectionFrom(tiles.get(0));
            }
            selectedTiles.addAll(tiles);
            lastTile = selectedTiles.get(selectedTiles.size() - 1);
        }

        //Builds letter to string to recognize it as a word
        @Override
        public String toString() {
            StringBuilder letters = new StringBuilder(selectedTiles.size());
            for (Tile letterTile : selectedTiles) {
                letters.append(((LetterTileView) letterTile.view).getLetter());
            }
            return letters.toString();
        }

        public BoardWord toBoardWord() {
            return new BoardWord(toString(), direction, getInitialTile());
        }

        @Override
        public int describeContents() {
            return 0;
        }

        @Override
        public void writeToParcel(Parcel dest, int flags) {
            dest.writeInt(direction);
            dest.writeParcelable(lastTile, flags);
            dest.writeTypedList(selectedTiles);
        }
    }

    public static class LetterTile extends Tile {

        private final String letter;

        public LetterTile(int row, int col, View view) {
            super(row, col, view);
            letter = ((LetterTileView)view).getLetter();
        }

        protected LetterTile(Parcel in) {
            super(in);
            letter = in.readString();
        }

        @Override
        public void writeToParcel(Parcel dest, int flags) {
            super.writeToParcel(dest, flags);
            dest.writeString(letter);
        }

        public static final Creator<LetterTile> CREATOR = new Creator<LetterTile>() {
            @Override
            public LetterTile createFromParcel(Parcel in) {
                return new LetterTile(in);
            }

            @Override
            public LetterTile[] newArray(int size) {
                return new LetterTile[size];
            }
        };
    }

    //CHANGED SAVEDSTATE to PUBLIC
    //Saves to Creator Key
    private static class SavedState extends BaseSavedState {

        private int boardRows, boardCols;
        private List<LetterTile> boardTiles;
        private List<SelectedWord> selectedWords;

        public SavedState(Parcelable superState) {
            super(superState);
        }

        private SavedState(Parcel in) {
            super(in);
            boardRows = in.readInt();
            boardCols = in.readInt();
            boardTiles = in.createTypedArrayList(LetterTile.CREATOR);
            selectedWords = in.createTypedArrayList(SelectedWord.CREATOR);
        }

        @Override
        public void writeToParcel(Parcel out, int flags) {
            super.writeToParcel(out, flags);
            out.writeInt(boardRows);
            out.writeInt(boardCols);
            out.writeTypedList(boardTiles);
            out.writeTypedList(selectedWords);
        }

        public static final Creator<SavedState> CREATOR = new Creator<SavedState>() {
            @Override
            public SavedState createFromParcel(Parcel in) {
                return new SavedState(in);
            }

            @Override
            public SavedState[] newArray(int size) {
                return new SavedState[size];
            }
        };
    }
}
